# coding=utf-8
from pyphysio.BaseAlgorithm import Algorithm
from abc import ABCMeta as _ABCMeta
__author__ = 'AleB'


class Indicator(Algorithm):
    """
    Algorithms that take as input a signal and return a scalar value.
    """
    __metaclass__ = _ABCMeta
    pass
